#!/bin/bash
#
#    emndl -- exponentially transformed Mandelbrot Set renderer
#    Copyright (C) 2011  Claude Heiland-Allen <claudiusmaximus@goto10.org>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

TIME0="$(date --iso=s)"
E="$(dirname "$(readlink -f "$0")")"
OUT="emndl_${TIME0}"

version()
{
cat << EOF
emndl.sh 0.1 (GPL) 2011  Claude Heiland-Allen <claudiusmaximus@goto10.org>
EOF
}

usage()
{
cat << EOF
usage options:
    --help -h -?    show this message and exit
    --version       show version and exit

input options:
    --auto bits     find an interesting point with precision 'bits'
    --re real       choose coordinates
    --im imag

render options:
    --quality q     horizontal size, (5 - 11)
    --depth d       vertical size, (2 - 24)

image options:
    --png           generate full size PNG
    --jpeg          generate thumbnail JPEG

video options:
    --aspect w:h    output aspect (4:3, 16:9)
    --dvd           generate PAL DVD (SD)
    --mkv           generate Matroska (HD)
    --ogv           generate Theora (LD)

other options:
    --quiet         less output
    --verbose       more output
EOF
}

if [ "x$*" == "x" ]
then
  version
  cat << EOF
Try \`emndl.sh --help' for usage information.
EOF
  exit 0
fi

TEMP=`getopt -o '' --long help,version,auto:,re:,im:,quality:,depth:,png,jpeg,aspect:,dvd,mkv,ogv,quiet,verbose -n 'emndl.sh' -- "$@"`
if [ $? != 0 ]
then
  echo "error 1 in getopt" >&2
  exit 1
fi

eval set -- "$TEMP"

AUTO=0
AUTOBITS="128"
REAL=0
REALCOORD="0.0"
IMAG=0
IMAGCOORD="0.0"
QUALITY="8"
DEPTH="16"
PNG=0
JPEG=0
ASPECT="16:9"
DVD=0
MKV=0
OGV=0
VERBOSE=1

while true
do
  case "$1" in
    --help|-h|-\?) usage ; exit 0 ;;
    --version) version ; exit 0 ;;
    --auto) AUTO=1 ; AUTOBITS="$2" ; shift 2 ;;
    --re) REAL=1 ; REALCOORD="$2" ; shift 2 ;;
    --im) IMAG=1 ; IMAGCOORD="$2" ; shift 2 ;;
    --quality) QUALITY="$2" ; shift 2 ;;
    --depth) DEPTH="$2" ; shift 2 ;;
    --png) PNG=1 ; shift ;;
    --jpeg) JPEG=1 ; shift ;;
    --aspect) ASPECT="$2" ; shift 2 ;;
    --dvd) DVD=1 ; shift ;;
    --mkv) MKV=1 ; shift ;;
    --ogv) OGV=1 ; shift ;;
    --quiet) VERBOSE=0 ; shift ;;
    --verbose) VERBOSE=2 ; shift ;;
    --) shift ; break ;;
    *) echo "error 2 in getopt" >&2 ; exit 1 ;;
  esac
done

(

case "$ASPECT" in
  4:3) WIDE=0 ;;
  16:9) WIDE=1 ;;
  *) echo "aspect $ASPECT not recognized" >&2 ; exit 1 ;;
esac

case "$QUALITY" in
  5) WIDTH=32 ;;
  6) WIDTH=64 ;;
  7) WIDTH=128 ;;
  8) WIDTH=256 ;;
  9) WIDTH=512 ;;
  10) WIDTH=1024 ;;
  11) WIDTH=2048 ;;
  *) echo "quality $QUALITY not recognized" >&2 ; exit 1 ;;
esac

if [ $AUTO -eq 1 ]
then
  # FIXME
  TEMP="$("${E}/emndl_autotune" +RTS -N -RTS "${AUTOBITS}")"
  REALCOORD="$(echo "${TEMP}" | sed 's|^\([^ ]*\) .*$|\1|')"
  IMAGCOORD="$(echo "${TEMP}" | sed 's|^[^ ]* \([^ ]*\) .*$|\1|')"
  DEPTH="$(echo "${TEMP}" | sed 's|^[^ ]* [^ ]* \([^ ]*\)$|\1|')"
fi
HEIGHT=$(( $WIDTH * $DEPTH ))

if [ $VERBOSE -gt 0 ]
then
  cat << EOF
AUTO=$AUTO
AUTOBITS=$AUTOBITS
REAL=$REAL
REALCOORD=$REALCOORD
IMAG=$IMAG
IMAGCOORD=$IMAGCOORD
QUALITY=$QUALITY
DEPTH=$DEPTH
PNG=$PNG
JPEG=$JPEG
ASPECT=$ASPECT
DVD=$DVD
MKV=$MKV
OGV=$OGV
VERBOSE=$VERBOSE

WIDE=$WIDE
WIDTH=$WIDTH
HEIGHT=$HEIGHT
E=$E
OUT=$OUT
EOF
fi

mkdir "${OUT}"
cd "${OUT}"

"${E}/emndl_parse" "${REALCOORD}" "${IMAGCOORD}" > "coords.qd"

if [ $VERBOSE -gt 0 ] ; then echo "initializing.." ; fi
"${E}/emndl_initialize" "0.emndl" "${WIDTH}" "${HEIGHT}" 1024 < coords.qd

if [ $VERBOSE -gt 0 ] ; then echo "iterating.." ; fi
j=0
for i in $(seq 0 36)
do
  j=$((i+1))
  if [ $VERBOSE -gt 0 ] ; then echo "iterating.. pass ${j}.." ; fi
  iters="$(echo "(5^${i})/(4^${i})*1000" | bc)"
  "${E}/emndl_iterate" "${iters}" "${i}.emndl" "${j}.emndl"
  if [ $? -eq 3 ] ; then break ; fi
done

if [ $VERBOSE -gt 0 ] ; then echo "finalizing.." ; fi
"${E}/emndl_finalize" "${j}.emndl" "o.f"

if [ $VERBOSE -gt 0 ] ; then echo "downscaling.." ; fi
case "${QUALITY}" in
  5)  cat "o.f" "/dev/zero" |
      head -c "$(( WIDTH * WIDTH * (DEPTH + 1) * 4 ))" |
      tee 00.f | "${E}/emndl_downscale"  32 |
      tee 01.f | "${E}/emndl_downscale"  16 |
      tee 02.f | "${E}/emndl_downscale"   8 |
      tee 03.f | "${E}/emndl_downscale"   4 |
      tee 04.f | "${E}/emndl_downscale"   2 |
      tee 05.f > /dev/null ;;
  6)  cat "o.f" "/dev/zero" |
      head -c "$(( WIDTH * WIDTH * (DEPTH + 1) * 4 ))" |
      tee 00.f | "${E}/emndl_downscale"  64 |
      tee 01.f | "${E}/emndl_downscale"  32 |
      tee 02.f | "${E}/emndl_downscale"  16 |
      tee 03.f | "${E}/emndl_downscale"   8 |
      tee 04.f | "${E}/emndl_downscale"   4 |
      tee 05.f | "${E}/emndl_downscale"   2 |
      tee 06.f > /dev/null ;;
  7)  cat "o.f" "/dev/zero" |
      head -c "$(( WIDTH * WIDTH * (DEPTH + 1) * 4 ))" |
      tee 00.f | "${E}/emndl_downscale" 128 |
      tee 01.f | "${E}/emndl_downscale"  64 |
      tee 02.f | "${E}/emndl_downscale"  32 |
      tee 03.f | "${E}/emndl_downscale"  16 |
      tee 04.f | "${E}/emndl_downscale"   8 |
      tee 05.f | "${E}/emndl_downscale"   4 |
      tee 06.f | "${E}/emndl_downscale"   2 |
      tee 07.f > /dev/null ;;
  8)  cat "o.f" "/dev/zero" |
      head -c "$(( WIDTH * WIDTH * (DEPTH + 1) * 4 ))" |
      tee 00.f | "${E}/emndl_downscale" 256 |
      tee 01.f | "${E}/emndl_downscale" 128 |
      tee 02.f | "${E}/emndl_downscale"  64 |
      tee 03.f | "${E}/emndl_downscale"  32 |
      tee 04.f | "${E}/emndl_downscale"  16 |
      tee 05.f | "${E}/emndl_downscale"   8 |
      tee 06.f | "${E}/emndl_downscale"   4 |
      tee 07.f | "${E}/emndl_downscale"   2 |
      tee 08.f > /dev/null ;;
  9)  cat "o.f" "/dev/zero" |
      head -c "$(( WIDTH * WIDTH * (DEPTH + 1) * 4 ))" |
      tee 00.f | "${E}/emndl_downscale"  512 |
      tee 01.f | "${E}/emndl_downscale"  256 |
      tee 02.f | "${E}/emndl_downscale"  128 |
      tee 03.f | "${E}/emndl_downscale"   64 |
      tee 04.f | "${E}/emndl_downscale"   32 |
      tee 05.f | "${E}/emndl_downscale"   16 |
      tee 06.f | "${E}/emndl_downscale"    8 |
      tee 07.f | "${E}/emndl_downscale"    4 |
      tee 08.f | "${E}/emndl_downscale"    2 |
      tee 09.f > /dev/null ;;
  10) cat "o.f" "/dev/zero" |
      head -c "$(( WIDTH * WIDTH * (DEPTH + 1) * 4 ))" |
      tee 00.f | "${E}/emndl_downscale" 1024 |
      tee 01.f | "${E}/emndl_downscale"  512 |
      tee 02.f | "${E}/emndl_downscale"  256 |
      tee 03.f | "${E}/emndl_downscale"  128 |
      tee 04.f | "${E}/emndl_downscale"   64 |
      tee 05.f | "${E}/emndl_downscale"   32 |
      tee 06.f | "${E}/emndl_downscale"   16 |
      tee 07.f | "${E}/emndl_downscale"    8 |
      tee 08.f | "${E}/emndl_downscale"    4 |
      tee 09.f | "${E}/emndl_downscale"    2 |
      tee 10.f > /dev/null ;;
  11) cat "o.f" "/dev/zero" |
      head -c "$(( WIDTH * WIDTH * (DEPTH + 1) * 4 ))" |
      tee 00.f | "${E}/emndl_downscale" 2048 |
      tee 01.f | "${E}/emndl_downscale" 1024 |
      tee 02.f | "${E}/emndl_downscale"  512 |
      tee 03.f | "${E}/emndl_downscale"  256 |
      tee 04.f | "${E}/emndl_downscale"  128 |
      tee 05.f | "${E}/emndl_downscale"   64 |
      tee 06.f | "${E}/emndl_downscale"   32 |
      tee 07.f | "${E}/emndl_downscale"   16 |
      tee 08.f | "${E}/emndl_downscale"    8 |
      tee 09.f | "${E}/emndl_downscale"    4 |
      tee 10.f | "${E}/emndl_downscale"    2 |
      tee 11.f > /dev/null ;;
  *)  echo "quality $QUALITY not fully supported yet" >&2 ; exit 1 ;;
esac

if [ $VERBOSE -gt 0 ] ; then echo "colourizing.." ; fi
for f in ??.f
do
  r="${f%f}raw"
  "${E}/emndl_colourize" "${f}" "${r}"
  s="$(stat -c '%s' "${r}")"
  w="$(echo "sqrt($s/3/(${DEPTH}+1))" | bc)"
  h="$(echo "$w*(${DEPTH}+1)" | bc)"
  ( echo -e "P6\n${w} ${h} 255" && cat "${r}" ) > "emndl.${f%f}ppm"
done

if [ $VERBOSE -gt 0 ] ; then echo "making images.." ; fi
if [ $PNG  -eq 1 ] ; then convert "emndl.00.ppm" "out.png" ; fi
if [ $JPEG -eq 1 ] ; then convert "emndl.00.ppm" -thumbnail 64x "out.jpeg" ; fi

if [ $VERBOSE -gt 0 ] ; then echo "making videos.." ; fi
if [ $DVD -eq 1 ] ; then mkfifo dvd.fifo ; ( cat dvd.fifo | y4mscaler -v 0 -O preset=dvd_wide -O yscale=1/1 | mpeg2enc -v 0 -f 8 -q 3 -b 8000 -B 768 -D 9 -g 9 -G 15 -P -R 2 -o "video.m2v" ) & fi
if [ $MKV -eq 1 ] ; then mkfifo mkv.fifo ; ( cat mkv.fifo | ffmpeg -f yuv4mpegpipe -b 2500000 -i - video.mkv ) & fi
if [ $OGV -eq 1 ] ; then mkfifo ogv.fifo ; ( cat ogv.fifo | ffmpeg2theora --aspect="${ASPECT}" -p preview -o video.ogv - >/dev/null 2>&1 ) & fi
if [ $((DVD + MKV + OGV)) -gt 0 ]
then
  "${E}/emndl_unwarp" emndl 1044 576 3 |
  "${E}/emndl_ppmtoy4m" 1044 576 |
  if [ $DVD -eq 1 ] ; then tee -a dvd.fifo ; else cat ; fi |
  if [ $MKV -eq 1 ] ; then tee -a mkv.fifo ; else cat ; fi |
  if [ $OGV -eq 1 ] ; then tee -a ogv.fifo ; else cat ; fi > /dev/null
fi

wait

cd ..
TIME1="$(date --iso=s)"

if [ $VERBOSE -gt 0 ]
then
  echo "${TIME0}  begin"
  echo "${TIME1}  end"
fi

) 2>&1 | tee "${OUT}.log"

exit 0
