/*
    emndl -- exponentially transformed Mandelbrot Set renderer
    Copyright (C) 2011  Claude Heiland-Allen <claudiusmaximus@goto10.org>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <cmath>
#include <qd/dd_real.h>
#include <qd/qd_real.h>
#include <qd/fpu.h>

#include "emndl.h"

float   to(const qd_real &x, const float   &dummy) { return to_double(x); }
double  to(const qd_real &x, const double  &dummy) { return to_double(x); }
dd_real to(const qd_real &x, const dd_real &dummy) { return to_dd_real(x); }
qd_real to(const qd_real &x, const qd_real &dummy) { return x; }

template<typename T>
void writeiters(int fdo, const qd_real &r, const qd_real &g, const qd_real &cx, const qd_real &cy, int w, const T &dummy) {
  int s = sizeof(struct iter<T>);
  if (sizeof(int) != write(fdo, &s, sizeof(int))) exit(1);
  for (int x = 0; x < w; ++x) {
    qd_real a(qd_real(x) * g);
    struct iter<T> i;
    i.cx = to(cx + r * cos(a), T(0.0));
    i.cy = to(cy + r * sin(a), T(0.0));
    i.zx = to(qd_real(0.0), T(0.0));
    i.zy = to(qd_real(0.0), T(0.0));
    i.dx = to(qd_real(0.0), T(0.0));
    i.dy = to(qd_real(0.0), T(0.0));
    i.psize = to(qd_real(4.0) / (r * g * qd_real::_log2 * sqrt(2)), T(0.0));
    i.n = 0;
    i.i = x;
    if (sizeof(struct iter<T>) != write(fdo, &i, sizeof(struct iter<T>))) exit(1);
  }
}

int main(int argc, char **argv) {
  fpu_fix_start(0);
  if (argc != 5) exit(1);
  int fdo = open(argv[1], O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
  int w = atoi(argv[2]);
  int h = atoi(argv[3]);
  qd_real r0(argv[4]);
  qd_real cx; if (1 != fread(&cx, sizeof(qd_real), 1, stdin)) exit(1);
  qd_real cy; if (1 != fread(&cy, sizeof(qd_real), 1, stdin)) exit(1);
  qd_real f(qd_real(8.0) / qd_real(w));
  qd_real g(qd_real(2) * qd_real::_pi / qd_real(w));
  for (int y = 0; y < h; ++y) {
    if (sizeof(int) != write(fdo, &w, sizeof(int))) exit(1);
    float d = -1;
    for (int x = 0; x < w; ++x) {
      if (sizeof(float) != write(fdo, &d, sizeof(float))) exit(1);
    }
    qd_real r(r0 * pow(qd_real(0.5), qd_real(y) * f));
    if (r > 1e-4) {
      writeiters(fdo, r, g, cx, cy, w, float(0.0));
    } else if (r > 1e-8) {
      writeiters(fdo, r, g, cx, cy, w, double(0.0));
    } else if (r > 1e-16) {
      writeiters(fdo, r, g, cx, cy, w, dd_real(0.0));
    } else { // if (r > 1e-48) {
      writeiters(fdo, r, g, cx, cy, w, qd_real(0.0));
    }
//    fprintf(stderr, "%5d\r", y+1);
  }
  close(fdo);
  return 0;
}
