/*
    emndl -- exponentially transformed Mandelbrot Set renderer
    Copyright (C) 2011  Claude Heiland-Allen <claudiusmaximus@goto10.org>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <cmath>
#include <qd/dd_real.h>
#include <qd/qd_real.h>
#include <qd/fpu.h>

#include "emndl.h"

#define likely(x)   __builtin_expect((x),1)
#define unlikely(x) __builtin_expect((x),0)

using namespace std;

float to_float(const float   &x) { return x; }
float to_float(const double  &x) { return x; }
float to_float(const dd_real &x) { return to_double(x); }
float to_float(const qd_real &x) { return to_double(x); }

template<typename T>
int iterate(float *dwell, struct iter<T> *iters, int k, int maxiters) {
  int progress = 0;
  T er2 = 65536;
  int p;
  #pragma omp parallel for private(p) reduction(+:progress) schedule(static, 1)
  for (int i = 0; i < k; ++i) {
    p = 0;
    if (iters[i].i < 0) continue;
    T zx = iters[i].zx;
    T zy = iters[i].zy;
    T dx = iters[i].dx;
    T dy = iters[i].dy;
    T cx = iters[i].cx;
    T cy = iters[i].cy;
    int n = iters[i].n;
    T zx2 = zx * zx;
    T zy2 = zy * zy;
    T z2 = zx2 + zy2;
    if (z2 < er2 && n < maxiters) {
//      int count = 1024;
      T z2xy = 2 * zx * zy;
//      while (likely(count-- && z2 < er2)) {
      while (likely(n < maxiters && z2 < er2)) {
        T zdzx = zx * dx - zy * dy;
        T zdzy = zx * dy + zy * dx;
        dx = 2 * zdzx + 1;
        dy = 2 * zdzy;
        zx = zx2 - zy2 + cx;
        zy = z2xy + cy;
        zx2 = zx * zx;
        zy2 = zy * zy;
        z2xy = 2 * zx * zy;
        z2 = zx2 + zy2;
        ++n;
      }
      iters[i].zx = zx;
      iters[i].zy = zy;
      iters[i].dx = dx;
      iters[i].dy = dy;
      iters[i].n = n;
      if (z2 >= er2) {
//        T d = n - log(log(z2) / log(er2)) / log(T(2));
//        dwell[iters[i].i] = to_float(log(d + 2));
        dwell[iters[i].i] = to_float(log(z2) * sqrt(z2 / (dx * dx + dy * dy)) * iters[i].psize);
        iters[i].i = -1;
        p = 1;
      }
    }
    progress = progress + p;
  }
  return progress;
}

template<typename T>
int writeiters(int fdo, struct iter<T> * iters, int k0) {
  int k = 0;
  for (int i = 0; i < k0; ++i) {
    if (iters[i].i >= 0) {
      if (sizeof(struct iter<T>) != write(fdo, &iters[i], sizeof(struct iter<T>))) exit(1);
      ++k;
    }
  }
  return k;
}

int main(int argc, char **argv) {
  fpu_fix_start(0);
  if (argc != 4) exit(1);
  int maxiters = atoi(argv[1]);
  int fdi = open(argv[2], O_RDONLY);
  int fdo = open(argv[3], O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
  int width;
  int oldwidth = 0;
  float *dwell = 0;
  int ibytes;
  int oldibytes = 0;
  void *iters = 0;
  int progress = 0;
  int y = 0;
  while (sizeof(int) == read(fdi, &width, sizeof(int))) {
    ++y;
    int dbytes = width * sizeof(float);
    if (width != oldwidth) {
      dwell = (float *) realloc(dwell, dbytes);
      oldwidth = width;
    }
    if (dbytes != read(fdi, dwell, dbytes)) exit(1);
    int k = 0;
    for (int i = 0; i < width; ++i) {
      if (dwell[i] < 0) ++k;
    }
    int prec;
    if (k > 0) {
      if (sizeof(int) != read(fdi, &prec, sizeof(int))) exit(1);
      ibytes = k * prec;
      if (ibytes > oldibytes) {
        iters = realloc(iters, ibytes);
        oldibytes = ibytes;
      }
      if (ibytes != read(fdi, iters, ibytes)) exit(1);
      int p = 0;
      switch (prec) {
#define CASE(T) \
case sizeof(struct iter<T>): \
  do { progress += p; } while ((p = iterate(dwell, (struct iter<T> *) iters, k, maxiters))); \
  break
        CASE(float);
        CASE(double);
        CASE(dd_real);
        CASE(qd_real);
#undef  CASE
        default:
          exit(1);
      }
    }
    if (sizeof(int) != write(fdo, &width, sizeof(int))) exit(1);
    if (dbytes != write(fdo, dwell, dbytes)) exit(1);
    int k0 = k;
    k = 0;
    for (int i = 0; i < width; ++i) {
      if (dwell[i] < 0) ++k;
    }
    if (k > 0) {
      if (sizeof(int) != write(fdo, &prec, sizeof(int))) exit(1);
      switch (prec) {
#define CASE(T) \
case sizeof(struct iter<T>): \
  if (k != writeiters(fdo, (struct iter<T> *) iters, k0)) exit(1); \
  break
        CASE(float);
        CASE(double);
        CASE(dd_real);
        CASE(qd_real);
#undef  CASE
        default:
          exit(1);
      }
    }
//    fprintf(stderr, "%5d\r", ++y);
  }
  int height = y;
  close(fdi);
  close(fdo);
  return ((!progress) || (progress > 0.001 * width * height)) ? 0 : 3;
}
